#==========================================================================
# iWidgets (experimental code!)
# Package version: 0.1-7
# (C)Copyright 2004-6 Simon Urbanek, All rights reserved.
#
# $Id: widgets.R 26 2007-07-26 00:19:02Z urbanek $
#==========================================================================

library(rJava)

inJGR <- FALSE

.First.lib <- function(libname, pkgname) {
    if (!require("rJava") || !require("JavaGD")) stop("iWidgets currently require rJava, JavaGD and JGR.")
    .jinit(paste(libname,pkgname,"java","iwidgets.jar",sep=.Platform$file.sep))

    iv<-try(.jcall("org/rosuda/iWidgets/Common","I","getIntVersion"),TRUE)
    if (!is.integer(iv))
        stop("Cannot find correct iWidgets classes.\nPlease note that currently iWidgets work only with JGR DP4 or later.")
    if (iv<256) stop("The iWidgets package is more recent than the iWidgets classes.\nYou may need to update JGR.")
    if (!exists(".getJavaGDObject"))
      stop("iWidgets require JavaGD 0.1-4 or later. Your JavaGD version is older.\nPlease update JavaGD at least to version 0.1-4. (see http://www.rosuda.org/R/)")
    add.classes <- character()
    if (is.jnull(.jfindClass("org/rosuda/JRI/REXP", silent = TRUE)))
      add.classes <- system.file("jri", "JRI.jar", package = "rJava")
    if (is.jnull(.jfindClass("org/rosuda/ibase/Common", silent = TRUE))) {
      n <- system.file("java", "iplots.jar", package = "iplots")
      if (nchar(n)) add.classes <- c(add.classes, n)
    }
    if (is.jnull(.jfindClass("org/rosuda/JGR/JGR", silent = TRUE))) {
      n <- system.file("java", "JGR.jar", package = "JGR")
      if (!nchar(n)) stop("JGR package is required (try install.packages('JGR'))")
      add.classes <- c(add.classes, n)
    }
    if (length(add.classes)) .jinit(paste(add.classes, collapse=.Platform$path.sep))
    inJGR <<- .jcall("org/rosuda/JGR/JGR", "Z", "isJGRmain", check=FALSE)
    .jcheck(silent=TRUE)
}

.handlers<-new.env()
.names<-new.env()
.root<-new.env()

.icomponent <- function(ref, class, name, handlers=NULL, window=NULL, ...) {
  o <- list(ref=ref, name=name, ...)
  class(o) <- c(class, "iComponent")
  if (!is.null(handlers))
    for (h in handlers) {
      d <- .jcall("org/rosuda/iWidgets/ActionDispatcher", "Lorg/rosuda/iWidgets/ActionDispatcher;","getGlobalDispatcher")
      r<-.jcall(d,"S","add",.jcast(o$ref,"java/lang/Object"))
      .jcall(o$ref,"V",h[[2]],.jcast(d,h[[3]]))
      h<-list(handler=h[[1]],name=r,obj=o)
      class(h)<-"iHandler"
      assign(r,h,envir=.handlers)
      o$handler.name=r
    }
  if (!is.null(name)) {
#    if (exists(name, .names))
#      warning("Component named `",name,"' already exists and will be repalced.")
    assign(name, o, .names)
  }
  if (!is.null(window)) add(window, o)
  o
}

wget <- function(name) {
  get(name, .names)
}

.icontainer <- function(ref, class, name, handlers=NULL, children=NULL, ...) {
  o <- .icomponent(ref, c(class, "iContainer"), name, handlers, NULL, ch=new.env(parent=emptyenv()), chn=new.env(parent=emptyenv()), ...)
  if (!is.null(children) & is.list(children) & length(children)>0)
    for (ch in children) add(o, ch)
  o
}

iwindow <- function(title="Window", visible=FALSE, horizontal=FALSE, handler=NULL, name=NULL, add=NULL, tag=NULL, ...) {
  f<-.jnew("org/rosuda/iWidgets/iWindow",as.character(title), as.logical(horizontal))
  o <- .icontainer(f, "iWindow", name, if (is.null(handler)) NULL else list(list(handler, "addWindowListener","java/awt/event/WindowListener")), add, ...)
  if (visible) visible(o, TRUE)
  o
}

igroup <- function(horizontal=TRUE, window=NULL, name=NULL, add=NULL, ...) {
  g <- .jnew("org/rosuda/iWidgets/iGroup", as.logical(horizontal))
  o <- .icontainer(g, "iGroup", name, NULL, add, ...)
  if (!is.null(window)) add(window, o)
  o
}

ilabel <- function(text, window=NULL, name=NULL, ...) {
  g <- .jnew("org/rosuda/iWidgets/iLabel", as.character(text))
  .icomponent(g, "iLabel", name, NULL, window, ...)
}

ilist <- function(content, handler=NULL, window=NULL, name=NULL, ...) {
  if(length(content)<2) stop("A list must have at least two elements")
  g <- .jnew("org/rosuda/iWidgets/iList", as.character(content))
  .icomponent(g, "iList", name, list(list(handler, "addListSelectionListener","javax/swing/event/ListSelectionListener")), window, ...)
}

idroplist <- function(content, editable=FALSE, handler=NULL, window=NULL, name=NULL, ...) {
  g <- .jnew("org/rosuda/iWidgets/iComboBox", as.character(content), as.logical(editable))
  .icomponent(g, "iDropList", name, list(list(handler, "addActionListener","java/awt/event/ActionListener")), window, ...)
}

ibutton <- function(text="OK", action=text, handler=NULL, window=NULL, name=NULL, ...) {
  b <- .jnew("org/rosuda/iWidgets/iButton", as.character(text), as.character(action))
  .icomponent(b, "iButton", name, list(list(handler, "addActionListener","java/awt/event/ActionListener")), window, text=text, action=action, ...)
}

iedit <- function(text="", width=10, action=NULL, handler=NULL, window=NULL, name=NULL, ...) {
  if (is.null(text)) text<-""
  b <- .jnew("org/rosuda/iWidgets/iEdit", as.character(text), as.integer(width))
  if (!is.null(action)) .jcall(b,"V","setActionCommand",as.character(action))
  .icomponent(b, "iEdit", name, list(list(handler, "addActionListener","java/awt/event/ActionListener")), window,
              text=text, width=width, action=action, ...)
}

icheckbox <- function(text, checked=FALSE, handler=NULL, window=NULL, name=NULL, ...) {
  b <- .jnew("org/rosuda/iWidgets/iCheckbox", as.character(text), as.logical(checked))
  .icomponent(b, "iCheckbox", name, list(list(handler, "addActionListener","java/awt/event/ActionListener")), window, text=text, ...)
}

iradio <- function(items, selected=1, horizontal=FALSE, handler=NULL, window=NULL, name=NULL, ...) {
  if (length(items)<2) stop("Radio button group makes sense only with at least two items.")
  b <- .jnew("org/rosuda/iWidgets/iRadio", as.character(items), as.integer(selected-1), as.logical(horizontal))
  .icomponent(b, "iRadio", name, list(list(handler, "addActionListener","java/awt/event/ActionListener")), window, items=items, ...)
}

islider <- function(min=0, max=100, value=min, horizontal=TRUE, handler=NULL, window=NULL, name=NULL, ...) {
    g <- .jnew("org/rosuda/iWidgets/iSlider", as.logical(horizontal), as.integer(min), as.integer(max), as.integer(value))
    .icomponent(g, "iSlider", name, list(list(handler, "addChangeListener","javax/swing/event/ChangeListener")), window, ...)
}

igraphics <- function(width=400, height=300, ps=12, window=NULL, name=NULL, ...) {
    a <- Sys.getenv("JAVAGD_CLASS_NAME")
    Sys.putenv("JAVAGD_CLASS_NAME"="org/rosuda/iWidgets/iGD")
    JavaGD(w=width, h=height, ps=ps)
    Sys.putenv("JAVAGD_CLASS_NAME"=a)
    di <- dev.cur()
    j <- .getJavaGDObject(di)
    if (is.null(j))
        stop("Unable to create Java graphics device.")
    c <- .jcall(j, "Ljavax/swing/JPanel;", "getPanel")
    if (is.null(c))
        c <- .jcall(j, "Ljava/awt/Canvas;", "getCanvas")
    if (is.null(c)) {
        dev.off()
        stop("Unable to get Java graphics content - if in doubt, please update both iWidgets and the JavaGD packages")
    }
    .icomponent(c, "iGD", name, NULL, window, gdobj=j, device=di, ...)
}

print.iComponent <- function(x,...) {
  cat(paste(class(x)[1],"[",.jstrVal(x$ref),"]\n",sep=''))
}

get.value <- function(...) UseMethod("get.value")
set.value <- function(...) UseMethod("set.value")

get.value.iSlider <- function(s)
    .jcall(s$ref, "I", "getValue")
set.value.iSlider <- function(s, value)
    invisible(.jcall(s$ref, "V", "setValue", as.integer(value)))

get.value.iLabel <- function(s)
    .jcall(s$ref, "S", "getText")
set.value.iLabel <- function(s, value)
    invisible(.jcall(s$ref, "V", "setText", as.character(value)))

get.value.iEdit <- function(s)
    .jcall(s$ref, "S", "getText")
set.value.iEdit <- function(s, value)
    invisible(.jcall(s$ref, "V", "setText", as.character(value)))

get.value.iRadio <- function(s)
    .jcall(s$ref, "I", "getSelectedID")+1
set.value.iRadio <- function(s, value)
    invisible(.jcall(s$ref, "V", "setSelectedID", as.integer(value-1)))

get.value.iCheckbox <- function(s)
    .jcall(s$ref, "Z", "isSelected")
set.value.iCheckbox <- function(s, value)
    invisible(.jcall(s$ref, "V", "setSelected", as.logical(value)))

get.value.iDropList <- function(s, index=FALSE)
    if (!index) .jcall(s$ref, "S", "getTextContent") else .jcall(s$ref, "I", "getSelectedIndex")+1
set.value.iDropList <- function(s, value, index=FALSE)
    invisible(if (!index) .jcall(s$ref, "V", "setTextContent", as.logical(TRUE)) else .jcall(s$ref, "V", "setSelectedIndex", as.integer(value-1)))


add.spring <- function(cont) {
    .jcall(cont$ref,"V","addGlue")
    invisible(NULL)
}

add.space <- function(cont, amount, horizontal=TRUE) {
    .jcall(cont$ref, "V", if(horizontal) "addHSpace" else "addVSpace", as.integer(amount))
    invisible(NULL)
}

add.strut <- add.space

add <- function(...) UseMethod("add")
delete <- function(...) UseMethod("delete")
visible <- function(...) UseMethod("visible")
"visible<-" <- function(...) UseMethod("visible")
dispose <- function(...) UseMethod("dispose")

#"[.iContainer"<-function(x,i) if (is.numeric(i)) { if (length(i)==1) x$ch[[as.character(i)]] else lapply(as.character(i),function(a) x$ch[[a]]) } else x$chn[[as.character(i)]]
#"[<-.iContainer"<-function(...) stop("Replacing components is not supported yet.")
#length.iContainer<-function() length(x$ch)

update.iComponent <- function(c) {
    .jcall(c$ref, "V", "validate")
}

update.iWindow <- function(c) {
    .jcall(c$ref, "V", "pack")
    .jcall(c$ref, "V", "validate")
}

dispose.iWindow <- function(win) {
    .jcall(win$ref,"V","dispose")
}

close.iWindow <- dispose.iWindow

add.iContainer <- function(cont, widget) {
  .jcall(cont$ref, "Ljava/awt/Component;", "add", .jcast(widget$ref, "java/awt/Component"))
  rid<-as.character(length(cont$ch)+1)
  assign(rid, widget, cont$ch)
  if (!is.null(widget$name))
    assign(widget$name,rid,cont$chn)
  if (is.null(getOption("auto-update")) || getOption("auto-update")) update(cont)
  invisible(NULL)
}

delete.iContainer <- function(cont, widget) {
    .jcall(cont$ref, "V", "remove",  .jcast(widget$ref, "java/awt/Component"))
    if (is.null(getOption("auto-update")) || getOption("auto-update")) update(cont)
    invisible(NULL)
}

visible.iComponent <- function(obj, set=NULL)
    if (is.null(set)) .jcall(obj$ref, "Z", "isVisible") else invisible(.jcall(obj$ref, "V", "setVisible", as.logical(set)))

.dispatch.event.for <- function(name,...) {
    h<-try(get(name,envir=.handlers),TRUE)
    if (inherits(h,"iHandler") && is.function(h$handler)) h$handler(h,...)
}
